import sqlite3
from collections import OrderedDict
import os
import shutil
import io
import importlib
import numpy as np
import codecs
import copy
import zipfile
import logging
import csv
from io import BytesIO
from tempfile import gettempdir
from ...reference_files import spatialite_database
from ...utils import WorkerThread

# from ...utils import WorkerThread
from ...parameters import Parameters
from .parse_csv import parse_csv

have_pyqt5 = importlib.util.find_spec("PyQt5")
if have_pyqt5 is None:
    pyqt = False

else:
    from PyQt5.QtCore import pyqtSignal as SIGNAL

    pyqt = True


# TODO : Add control for mandatory and optional files
# TODO: Add constraints for non-negative and limited options (0,1,2, etc) for fields through foreign keys
class create_gtfsdb(WorkerThread):
    converting_gtfs = SIGNAL(object)

    def __init__(self, file_path, save_db, memory_db=False, spatialite_enabled=False, overwrite=False):
        WorkerThread.__init__(self, None)
        self.conn = None
        self.cursor = None
        self.__max_chunk_size = None
        self.spatialite_enabled = spatialite_enabled
        self.available_files = {}
        self.source_path = file_path
        self.save_db = save_db
        self.memory_db = memory_db
        self.overwrite = overwrite
        self.report = []
        self.logger = logging.getLogger("aequilibrae")
        log_level = Parameters().parameters["system"]["logging"]
        if isinstance(log_level, str):
            log_level = log_level.upper()
        self.logger.setLevel(log_level)

        self.logger.info("Starting GTFS import")
        self.logger.info("      " + self.source_path)
        self.logger.info("      " + str(self.save_db))

        if os.path.isfile(self.source_path):
            zip_container = zipfile.ZipFile(self.source_path)
            ret = zip_container.testzip()
            if ret is not None:
                raise ValueError("First bad file in zip: %s" % ret)
            else:
                self.source = "zip"
        elif os.path.isdir(self.source_path):
            self.source = "folder"
        else:
            msg = "Source needs to be zip file or directory/folder"
            self.logger.error(msg)
            raise ValueError(msg)

        # Input consistency checks
        if self.source is None:
            msg = "Input is neither a zip file nor a folder"
            self.logger.error("      " + msg)
            raise ValueError(msg)

        if spatialite_enabled and memory_db:
            msg = "Spatialite is only supported on disk"
            self.logger.error("      " + msg)
            raise ValueError(msg)

        if not self.overwrite:
            if os.path.isfile(os.path.join(save_db)):
                msg = "Output database exists. Please use overwrite=True or choose a different path/name"
                self.logger.error("      " + msg)
                raise ValueError(msg)
        else:
            if os.path.isfile(save_db):
                os.unlink(save_db)

        if self.save_db is None:
            if not self.memory_db:
                save_db = ":memory:"
        else:
            if self.memory_db:
                msg = "You can't have a file name and have the file in memory at the same time"
                self.logger.error("      " + msg)
                raise ValueError(msg)

        OrderedDict([("s", (1, 2)), ("p", (3, 4)), ("a", (5, 6)), ("m", (7, 8))])
        self.column_order = {
            "agency.txt": OrderedDict(
                [
                    ("agency_id", str),
                    ("agency_name", str),
                    ("agency_url", str),
                    ("agency_timezone", str),
                    ("agency_lang", str),
                    ("agency_phone", str),
                    ("agency_fare_url", str),
                    ("agency_email", str),
                ]
            ),
            "routes.txt": OrderedDict(
                [
                    ("route_id", str),
                    ("agency_id", str),
                    ("route_short_name", str),
                    ("route_long_name", str),
                    ("route_desc", str),
                    ("route_type", int),
                    ("route_url", str),
                    ("route_color", str),
                    ("route_text_color", str),
                    ("route_sort_order", int),
                ]
            ),
            "trips.txt": OrderedDict(
                [
                    ("route_id", str),
                    ("service_id", str),
                    ("trip_id", str),
                    ("trip_headsign", str),
                    ("trip_short_name", str),
                    ("direction_id", int),
                    ("block_id", str),
                    ("shape_id", str),
                    ("wheelchair_accessible", int),
                    ("bikes_allowed", int),
                ]
            ),
            "stop_times.txt": OrderedDict(
                [
                    ("trip_id", str),
                    ("arrival_time", str),
                    ("departure_time", str),
                    ("stop_id", str),
                    ("stop_sequence", int),
                    ("stop_headsign", str),
                    ("pickup_type", int),
                    ("shape_dist_traveled", float),
                    ("timepoint", int),
                ]
            ),
            "calendar.txt": OrderedDict(
                [
                    ("service_id", str),
                    ("monday", int),
                    ("tuesday", int),
                    ("wednesday", int),
                    ("thursday", int),
                    ("friday", int),
                    ("saturday", int),
                    ("sunday", int),
                    ("start_date", str),
                    ("end_date", str),
                ]
            ),
            "calendar_dates.txt": OrderedDict([("service_id", str), ("date", str), ("exception_type", int)]),
            "fare_attributes.txt": OrderedDict(
                [
                    ("fare_id", str),
                    ("price", float),
                    ("currency_type", str),
                    ("payment_method", int),
                    ("transfers", int),
                    ("agency_id", str),
                    ("transfer_duration", float),
                ]
            ),
            "fare_rules.txt": OrderedDict(
                [("fare_id", str), ("route_id", str), ("origin_id", str), ("destination_id", str), ("contains_id", str)]
            ),
            "frequencies.txt": OrderedDict(
                [("trip_id", str), ("start_time", str), ("end_time", str), ("headway_secs", str), ("exact_times", int)]
            ),
            "transfers.txt": OrderedDict(
                [("from_stop_id", str), ("to_stop_id", str), ("transfer_type", int), ("min_transfer_time", int)]
            ),
            "feed_info.txt": OrderedDict(
                [
                    ("feed_publisher_name", str),
                    ("feed_publisher_url", str),
                    ("feed_lang", str),
                    ("feed_start_date", str),
                    ("feed_end_date", str),
                    ("feed_version", str),
                ]
            ),
            "stops.txt": OrderedDict(
                [
                    ("stop_id", str),
                    ("stop_code", str),
                    ("stop_name", str),
                    ("stop_desc", str),
                    ("stop_lat", float),
                    ("stop_lon", float),
                    ("zone_id", str),
                    ("stop_url", str),
                    ("location_type", int),
                    ("parent_station", str),
                    ("stop_timezone", str),
                    ("wheelchair_boarding", int),
                ]
            ),
            "shapes.txt": OrderedDict(
                [
                    ("shape_id", str),
                    ("shape_pt_lat", float),
                    ("shape_pt_lon", float),
                    ("shape_pt_sequence", int),
                    ("shape_dist_traveled", float),
                ]
            ),
        }
        self.set_chunk_size(30000)
        self.logger.info("      No errors found in initial checking")
        self.logger.info("      source: " + self.source)

    def set_chunk_size(self, chunk_size):
        if chunk_size is not None:
            if isinstance(chunk_size, int):
                self.__max_chunk_size = chunk_size

    def import_gtfs(self):
        # In case we have not create the database yet
        if self.conn is None:
            if pyqt:
                self.converting_gtfs.emit(["text", "Creating container database"])
            self.create_database()

        # Import all tables to SQLITE
        tables = [x.split(".")[0] for x in self.column_order.keys()]
        if self.spatialite_enabled:
            self.converting_gtfs.emit(["total files", 16])
        else:
            self.converting_gtfs.emit(["total files", 14])

        for i, tbl in enumerate(tables):
            if pyqt:
                self.converting_gtfs.emit(["text", "Loading data from file: " + tbl])
            self.logger.info("      loading table " + tbl)
            self.__load_tables(tbl)
            if pyqt:
                self.converting_gtfs.emit(["files counter", i + 1])

        # Creates the geometry
        if self.spatialite_enabled:
            self.__create_geometry()
        self.conn.commit()
        self.conn.close()
        self.converting_gtfs.emit(["finished_threaded_procedure", 0])

    def create_database(self):
        if pyqt:
            self.converting_gtfs.emit(["text", "Creating empty database"])
            # self.converting_gtfs.emit(SIGNAL("converting_gtfs"), ['text', 'Creating empty database'])

        if self.spatialite_enabled:
            shutil.copy(spatialite_database, self.save_db)

        self.conn = sqlite3.connect(self.save_db)
        self.cursor = self.conn.cursor()

        self.__create_empty_tables()

    def __create_empty_tables(self):
        self.__create_agency_table()
        self.__create_route_table()
        self.__create_trips_table()
        self.__create_stop_times_table()
        self.__create_calendar_table()
        self.__create_calendar_dates_table()
        self.__create_fare_attributes_table()
        self.__create_fare_rules_table()
        self.__create_frequencies_table()
        self.__create_transfers_table()
        self.__create_feed_info_table()
        self.__create_stops_table()
        self.__create_shapes_table()
        self.conn.commit()

    def __create_agency_table(self):
        self.cursor.execute("DROP TABLE IF EXISTS agency")
        create_query = """CREATE TABLE 'agency' (agency_id VARCHAR PRIMARY KEY UNIQUE,
                                                 agency_name VARCHAR  NOT NULL,
                                                 agency_url VARCHAR  NOT NULL,
                                                 agency_timezone VARCHAR  NOT NULL,
                                                 agency_lang VARCHAR,
                                                 agency_phone VARCHAR,
                                                 agency_fare_url VARCHAR,
                                                 agency_email VARCHAR);"""
        self.cursor.execute(create_query)

    def __create_route_table(self):
        self.cursor.execute("DROP TABLE IF EXISTS routes")
        create_query = """CREATE TABLE 'routes' (route_id VARCHAR PRIMARY KEY UNIQUE NOT NULL,
                                                 agency_id VARCHAR,
                                                 route_short_name VARCHAR,
                                                 route_long_name VARCHAR,
                                                 route_desc VARCHAR,
                                                 route_type NUMERIC NOT NULL,
                                                 route_url VARCHAR,
                                                 route_color VARCHAR,
                                                 route_text_color VARCHAR,
                                                 route_sort_order NUMERIC,
                                                 FOREIGN KEY(agency_id) REFERENCES agency(agency_id));"""

        self.cursor.execute(create_query)

    def __create_trips_table(self):
        self.cursor.execute("DROP TABLE IF EXISTS trips")
        # TODO: Add foreign key to calendar_dates.txt
        create_query = """CREATE TABLE 'trips' (route_id VARCHAR NOT NULL,
                                                service_id VARCHAR NOT NULL,
                                                trip_id VARCHAR PRIMARY KEY UNIQUE NOT NULL,
                                                trip_headsign VARCHAR,
                                                trip_short_name VARCHAR,
                                                direction_id NUMERIC,
                                                block_id VARCHAR,
                                                shape_id VARCHAR,
                                                wheelchair_accessible NUMERIC,
                                                bikes_allowed NUMERIC,
                                                FOREIGN KEY(route_id) REFERENCES routes(route_id)
                                                FOREIGN KEY(service_id) REFERENCES calendar(service_id));"""

        self.cursor.execute(create_query)

    def __create_calendar_table(self):
        self.cursor.execute("DROP TABLE IF EXISTS calendar")
        create_query = """CREATE TABLE 'calendar' (service_id VARCHAR NOT NULL,
                                                   monday NUMERIC NOT NULL DEFAULT 1,
                                                   tuesday NUMERIC NOT NULL DEFAULT 1,
                                                   wednesday NUMERIC NOT NULL DEFAULT 1,
                                                   thursday NUMERIC NOT NULL DEFAULT 1,
                                                   friday NUMERIC NOT NULL DEFAULT 1,
                                                   saturday NUMERIC NOT NULL DEFAULT 1,
                                                   sunday NUMERIC NOT NULL DEFAULT 1,
                                                   start_date VARCHAR,
                                                   end_date VARCHAR,
                                                   FOREIGN KEY(service_id) REFERENCES trips(service_id));"""

        self.cursor.execute(create_query)

    def __create_calendar_dates_table(self):
        # TODO: Add foreign key to calendar.txt
        self.cursor.execute("DROP TABLE IF EXISTS calendar_dates")
        create_query = """CREATE TABLE 'calendar_dates' (service_id VARCHAR NOT NULL,
                                                   'date' VARCHAR NOT NULL,
                                                   exception_type NUMERIC NOT NULL,
                                                   FOREIGN KEY(service_id) REFERENCES trips(service_id));"""
        self.cursor.execute(create_query)

    def __create_stop_times_table(self):
        self.cursor.execute("DROP TABLE IF EXISTS stop_times")
        create_query = """CREATE TABLE 'stop_times' (stop_time_id INTEGER PRIMARY KEY NOT NULL,
                                                     trip_id VARCHAR NOT NULL,
                                                     arrival_time VARCHAR NOT NULL,
                                                     departure_time VARCHAR NOT NULL,
                                                     stop_id VARCHAR NOT NULL,
                                                     stop_sequence NUMERIC NOT NULL,
                                                     stop_headsign VARCHAR,
                                                     pickup_type NUMERIC DEFAULT 0,
                                                     drop_off_type NUMERIC  DEFAULT 0,
                                                     shape_dist_traveled NUMERIC,
                                                     timepoint NUMERIC,
                                                     FOREIGN KEY(trip_id) REFERENCES trips(trip_id)
                                                     FOREIGN KEY(stop_id) REFERENCES stops(stop_id));"""

        self.cursor.execute(create_query)

    def __create_fare_attributes_table(self):
        self.cursor.execute("DROP TABLE IF EXISTS fare_attributes")
        create_query = """CREATE TABLE 'fare_attributes' (fare_id VARCHAR NOT NULL,
                                                          price NUMERIC NOT NULL,
                                                          currency_type VARCHAR NOT NULL,
                                                          payment_method NUMERIC NOT NULL,
                                                          transfers NUMERIC,
                                                          agency_id VARCHAR,
                                                          transfer_duration NUMERIC);"""

        self.cursor.execute(create_query)

    def __create_fare_rules_table(self):
        self.cursor.execute("DROP TABLE IF EXISTS fare_rules")
        create_query = """CREATE TABLE 'fare_rules' (fare_id VARCHAR NOT NULL,
                                                     route_id VARCHAR,
                                                     origin_id VARCHAR,
                                                     destination_id VARCHAR,
                                                     contains_id VARCHAR,
                                                     FOREIGN KEY(fare_id) REFERENCES fare_attributes(fare_id)
                                                     FOREIGN KEY(route_id) REFERENCES routes(route_id)
                                                     FOREIGN KEY(origin_id) REFERENCES stops(stop_id)
                                                     FOREIGN KEY(contains_id) REFERENCES stops(stop_id)
                                                     FOREIGN KEY(destination_id) REFERENCES stops(stop_id));"""
        self.cursor.execute(create_query)

    def __create_frequencies_table(self):
        self.cursor.execute("DROP TABLE IF EXISTS frequencies")
        create_query = """CREATE TABLE 'frequencies' (trip_id VARCHAR NOT NULL,
                                                     start_time VARCHAR NOT NULL,
                                                     end_time VARCHAR NOT NULL,
                                                     headway_secs VARCHAR NOT NULL,
                                                     exact_times NUMERIC,
                                                     FOREIGN KEY(trip_id) REFERENCES trips(trip_id));"""
        self.cursor.execute(create_query)

    def __create_transfers_table(self):
        self.cursor.execute("DROP TABLE IF EXISTS transfers")
        create_query = """CREATE TABLE 'transfers' (from_stop_id VARCHAR NOT NULL,
                                                    to_stop_id VARCHAR NOT NULL,
                                                    transfer_type NUMERIC NOT NULL,
                                                    min_transfer_time NUMERIC NOT NULL,
                                                    FOREIGN KEY(from_stop_id) REFERENCES stops(stop_id)
                                                    FOREIGN KEY(to_stop_id) REFERENCES stops(stop_id));"""
        self.cursor.execute(create_query)

    def __create_feed_info_table(self):
        self.cursor.execute("DROP TABLE IF EXISTS feed_info")
        create_query = """CREATE TABLE 'feed_info' (feed_publisher_name VARCHAR NOT NULL,
                                                    feed_publisher_url VARCHAR NOT NULL,
                                                    feed_lang VARCHAR NOT NULL,
                                                    feed_start_date VARCHAR,
                                                    feed_end_date VARCHAR,
                                                    feed_version VARCHAR);"""
        self.cursor.execute(create_query)

    def __create_stops_table(self):
        self.cursor.execute("DROP TABLE IF EXISTS stops")
        create_query = """CREATE TABLE 'stops' (stop_id VARCHAR PRIMARY KEY UNIQUE NOT NULL,
                                                stop_code VARCHAR,
                                                stop_name VARCHAR NOT NULL,
                                                stop_desc VARCHAR,
                                                stop_lat NUMERIC NOT NULL,
                                                stop_lon NUMERIC NOT NULL,
                                                zone_id VARCHAR,
                                                stop_url VARCHAR,
                                                location_type NUMERIC,
                                                parent_station VARCHAR,
                                                stop_timezone VARCHAR,
                                                wheelchair_boarding NUMERIC);"""
        self.cursor.execute(create_query)

    def __create_shapes_table(self):
        self.cursor.execute("DROP TABLE IF EXISTS shapes")
        create_query = """CREATE TABLE 'shapes' (shape_id VARCHAR NOT NULL,
                                                 shape_pt_lat NUMERIC NOT NULL,
                                                 shape_pt_lon NUMERIC NOT NULL,
                                                 shape_pt_sequence NUMERIC NOT NULL,
                                                 shape_dist_traveled NUMERIC);"""
        self.cursor.execute(create_query)

    def __load_tables(self, table_name):

        # list fields from table
        cursor = self.cursor.execute("select * from " + table_name)
        available_columns = [description[0].lower() for description in cursor.description]

        # create the file name
        file_to_open = table_name + ".txt"

        # Check if exists
        if self.source == "folder":
            data_file = os.path.join(self.source_path, file_to_open)
            self.logger.warning("          " + data_file)
            if not os.path.isfile(data_file):
                self.available_files[file_to_open] = False
                self.logger.warning("          Table " + table_name + " not available")
                return
            else:
                data_file = open(data_file, "r")
        else:
            zip_container = zipfile.ZipFile(self.source_path, "r")
            if file_to_open in zip_container.namelist():
                data_file = zip_container.open(file_to_open, "r")
                data_file = io.TextIOWrapper(data_file, encoding="utf-8-sig")
            else:
                self.available_files[file_to_open] = False
                self.logger.warning("          Table " + table_name + " not available")
                return

        self.available_files[file_to_open] = True
        data = parse_csv(data_file, column_order=self.column_order[file_to_open])

        if not isinstance(data_file, str):
            data_file.close()

        if data is None:
            return
        # we check which columns in the table structure are available in the dataset
        correspondence = []
        for col in data.dtype.names:
            if col in available_columns:
                correspondence.append(col)
        data = data[correspondence]
        cols = data.dtype.names
        dt = tuple(data.tolist())
        fields = ",".join(len(cols) * ["?"])
        try:
            if not isinstance(dt[0], tuple):
                dt = [dt]
            self.cursor.executemany(
                "INSERT into " + table_name + " (" + ",".join(cols) + ") VALUES(" + fields + ")", dt
            )
            self.conn.commit()
        except Exception as e:
            self.logger.error("Could not load data from table " + table_name)
            self.logger.error(str(e))

    def __create_geometry(self):
        # enable extension loading
        self.conn.enable_load_extension(True)
        self.cursor.execute("SELECT load_extension('mod_spatialite')")
        self.conn.commit()
        # We need to create three things here:
        # 1. A geometry column in the stops table
        # 2. A layer of shapes corresponding to each trip
        # 3. A layer of shapes with all the stops for each trip that can be query'd by route_id

        # 1
        if pyqt:
            self.converting_gtfs.emit(["text", "Creating stops' geometry"])
            self.converting_gtfs.emit(["files counter", 14])

        self.cursor.execute("SELECT AddGeometryColumn( 'stops', 'geometry', 4326, 'POINT', 'XY' );")
        self.cursor.execute("update stops set geometry=MakePoint(stop_lon ,stop_lat, 4326);")
        self.cursor.execute("SELECT CreateSpatialIndex( 'stops' , 'geometry' );")

        # 2
        if pyqt:
            self.converting_gtfs.emit(["text", "Creating routes' geometry"])
            self.converting_gtfs.emit(["files counter", 15])
        # We create the table to hold the shapes for each route
        self.cursor.execute("DROP TABLE IF EXISTS shape_routes")
        # TODO: Add foreign key to calendar_dates.txt
        create_query = """CREATE TABLE 'shape_routes' (route_id VARCHAR,
                                                       trip_id VARCHAR,
                                                       shape_id VARCHAR,
                                                       route_text_color VARCHAR,
                                                       FOREIGN KEY(route_id) REFERENCES routes(route_id)
                                                       FOREIGN KEY(trip_id) REFERENCES trips(trip_id));"""
        self.cursor.execute(create_query)
        self.cursor.execute("SELECT AddGeometryColumn( 'shape_routes', 'geometry', 4326, 'LINESTRING', 'XY' );")
        self.cursor.execute("SELECT CreateSpatialIndex( 'shape_routes' , 'geometry' );")

        # We check if we have shapes in the shape layer
        shape_ids = self.cursor.execute("SELECT DISTINCT shape_id from shapes;").fetchall()
        shape_ids = [str(x[0], "utf-8") for x in shape_ids]
        if len(shape_ids) > 0:
            if pyqt:
                self.converting_gtfs.emit(["max chunk counter", len(shape_ids)])

            for i, shp in enumerate(shape_ids):
                if pyqt:
                    self.converting_gtfs.emit(["chunk counter", i])

                qry = self.cursor.execute("SELECT route_id, trip_id from trips where shape_id='" + shp + "'").fetchall()
                if len(qry) > 0:
                    route_id, trip_id = qry[0]

                    points = self.cursor.execute(
                        "SELECT shape_pt_lon, shape_pt_lat from shapes where shape_id='"
                        + str(shp)
                        + "' order by shape_pt_sequence"
                    ).fetchall()
                    txt = "LINESTRING (" + ", ".join([str(x[0]) + " " + str(x[1]) for x in points]) + ")"
                    route_text_color = self.cursor.execute(
                        "SELECT route_text_color from routes where route_id='" + str(route_id) + "'"
                    ).fetchall()
                    if len(route_text_color):
                        route_text_color = route_text_color[0]
                        if len(route_text_color) > 0:
                            sql = (
                                "INSERT INTO shape_routes (route_id, trip_id, shape_id, route_text_color, geometry)"
                                + "VALUES (?,?,?,?,"
                                + "LineFromText('"
                                + str(txt)
                                + "', 4326))"
                            )
                            self.cursor.execute(sql, (route_id, trip_id, shp, route_text_color[0]))
        else:
            trip_ids = self.cursor.execute("SELECT DISTINCT trip_id from trips;").fetchall()
            trip_ids = [str(x[0]) for x in trip_ids]
            if pyqt:
                self.converting_gtfs.emit(["max chunk counter", len(trip_ids)])

            for i, trip_id in enumerate(trip_ids):
                if pyqt:
                    self.converting_gtfs.emit(["chunk counter", i])

                route_id = self.cursor.execute(
                    "SELECT route_id from trips where trip_id='" + str(trip_id) + "'"
                ).fetchone()[0]

                sql = (
                    "SELECT stop_lon, stop_lat FROM stop_times INNER JOIN stops"
                    " ON stop_times.stop_id = stops.stop_id WHERE stop_times.trip_id='"
                    + str(trip_id)
                    + "' order by stop_times.stop_sequence"
                )
                qry = self.cursor.execute(sql).fetchall()
                txt = "LINESTRING (" + ", ".join([str(x[0]) + " " + str(x[1]) for x in qry]) + ")"
                route_text_color = self.cursor.execute(
                    "SELECT route_text_color from routes where route_id='" + str(route_id) + "'"
                ).fetchall()[0]
                sql = (
                    "INSERT INTO shape_routes (route_id, trip_id, route_text_color, geometry) "
                    "VALUES (?,?," + "LineFromText('" + txt + "', 4326))"
                )
                self.cursor.execute(sql, (route_id, trip_id, route_text_color))

        # creates the stops table with route ID info
        if pyqt:
            self.converting_gtfs.emit(["text", "Associating routes and stops. Sit tight"])

        self.cursor.execute(
            """CREATE TABLE 'shape_stops'
                                AS
                                SELECT
                                a.stop_time_id,
                                a.trip_id,
                                a.stop_id,
                                c.route_id,
                                b.stop_lon,
                                b.stop_lat
                                FROM
                                stop_times a
                                inner join
                                stops  b
                                on a.stop_id  = b.stop_id
                                inner join
                                trips c
                                on a.trip_id = c.trip_id;"""
        )

        self.cursor.execute("SELECT AddGeometryColumn( 'shape_stops', 'geometry', 4326, 'POINT', 'XY' );")
        self.cursor.execute("update shape_stops set geometry=MakePoint(stop_lon ,stop_lat, 4326);")
        self.cursor.execute("SELECT CreateSpatialIndex( 'shape_stops' , 'geometry' );")
        self.conn.commit()



# cimport numpy as np
# cimport cython
# include 'parameters.pxi'
import numpy as np
import multiprocessing as mp
from ...matrix import AequilibraeMatrix


class SkimResults:
    def __init__(self):
        """
        @type graph: Set of numpy arrays to store Computation results
        """
        self.skims = None
        self.path = None
        self.path_nodes = None
        self.milepost = None
        self.cores = mp.cpu_count()

        self.links = -1
        self.nodes = -1
        self.zones = -1
        self.num_skims = -1
        self.__graph_id__ = None

    def prepare(self, graph):

        self.nodes = graph.num_nodes + 1
        self.zones = graph.num_zones
        self.links = graph.num_links + 1
        self.num_skims = len(graph.skim_fields)

        self.skims = AequilibraeMatrix()
        self.skims.create_empty(
            file_name=AequilibraeMatrix().random_name(),
            zones=self.zones,
            matrix_names=graph.skim_fields,
        )
        self.skims.index[:] = graph.centroids[:]
        self.skims.computational_view(core_list=self.skims.names)
        self.skims.matrix_view = self.skims.matrix_view.reshape(
            self.zones, self.zones, self.num_skims
        )
        self.__graph_id__ = graph.__id__

    def set_cores(self, cores):
        if isinstance(cores, int):
            if cores > 0:
                if self.cores != cores:
                    self.cores = cores
            else:
                raise ValueError("Number of cores needs to be equal or bigger than one")
        else:
            raise ValueError("Number of cores needs to be an integer")

    def reset(self):
        if self.skims is not None:
            self.skims.fill(np.inf)
            self.path = None
            self.path_nodes = None
            self.milepost = None

        else:
            raise ValueError(
                "Exception: Path results object was not yet prepared/initialized"
            )



from unittest import TestCase

from aequilibrae.transit.gtfs import create_gtfsdb
from ...data import gtfs_folder
from ...data import gtfs_zip
from ...data import gtfs_db_output


class TestCreate_gtfsdb(TestCase):
    def test_create_database(self):
        # TODO: Fix Travis-ci configuration in order to properly run the test with spatialite enabled
        self.gtfs = create_gtfsdb(gtfs_zip, save_db=gtfs_db_output, overwrite=True, spatialite_enabled=False)
        self.gtfs.create_database()
        self.gtfs.conn.close()

    def test_import_gtfs(self):
        # self.gtfs = create_gtfsdb(gtfs_folder, save_db=gtfs_db_output, overwrite=True, spatialite_enabled=True)
        self.gtfs = create_gtfsdb(gtfs_folder, save_db=gtfs_db_output, overwrite=True, spatialite_enabled=False)
        self.gtfs.import_gtfs()

        self.gtfs = create_gtfsdb(gtfs_zip, save_db=gtfs_db_output, overwrite=True, spatialite_enabled=False)
        self.gtfs.import_gtfs()



class Trip:
    def __init__(self):
        # Some characteristics are not needed because they will be inherited from a parent class
        # self.route_id

        # Others are needed
        self.service_id = None
        self.id = None
        self.head_sign = None
        self.short_name = None
        self.direction_id = None
        self.block_id = None
        self.shape_id = None
        self.wheelchair_accessible = None
        self.bikes_allowed = None



import numpy as np
from ..graph import Graph
from ..AoN import update_path_trace


class PathResults:
    def __init__(self):
        """
        @type graph: Set of numpy arrays to store Computation results
        """
        self.predecessors = None
        self.connectors = None
        self.skims = None
        self.path = None
        self.path_nodes = None
        self.milepost = None
        self.reached_first = None
        self.origin = None
        self.destination = None

        self.links = -1
        self.nodes = -1
        self.zones = -1
        self.num_skims = -1
        self.__integer_type = None
        self.__float_type = None
        self.__graph_id__ = None

    def prepare(self, graph):
        self.__integer_type = graph.default_types("int")
        self.__float_type = graph.default_types("float")
        self.nodes = graph.num_nodes + 1
        self.zones = graph.centroids + 1
        self.links = graph.num_links + 1
        self.num_skims = graph.skims.shape[1]

        self.predecessors = np.zeros(self.nodes, dtype=self.__integer_type)
        self.connectors = np.zeros(self.nodes, dtype=self.__integer_type)
        self.reached_first = np.zeros(self.nodes, dtype=self.__integer_type)
        self.skims = np.zeros((self.nodes, self.num_skims), self.__float_type)
        self.__graph_id__ = graph.__id__

    def reset(self):
        if self.predecessors is not None:
            self.predecessors.fill(-1)
            self.connectors.fill(-1)
            self.skims.fill(np.inf)
            self.path = None
            self.path_nodes = None
            self.milepost = None

        else:
            raise ValueError(
                "Exception: Path results object was not yet prepared/initialized"
            )

    def update_trace(self, graph, destination):
        # type: (Graph, int) -> (None)
        if not isinstance(destination, int):
            raise TypeError("destination needs to be an integer")

        if not isinstance(graph, Graph):
            raise TypeError("graph needs to be an AequilibraE Graph")

        if destination >= graph.nodes_to_indices.shape[0]:
            raise ValueError(
                "destination out of the range of node numbers in the graph"
            )

        if self.__graph_id__ != graph.__id__:
            raise ValueError(
                "Results object not prepared. Use --> results.prepare(graph)"
            )

        update_path_trace(self, destination, graph)



"""
 -----------------------------------------------------------------------------------------------------------
 Package:    AequilibraE

 Name:       Auxiliary code
 Purpose:    Compiles AequilibraE's Cython code

 Original Author:  Pedro Camargo (c@margo.co)
 Contributors:
 Last edited by: Pedro Camrgo

 Website:    www.AequilibraE.com
 Repository:  https://github.com/AequilibraE/AequilibraE

 Created:    01/01/2013
 Updated:    30/09/2016
 Copyright:   (c) AequilibraE authors
 Licence:     See LICENSE.TXT
 -----------------------------------------------------------------------------------------------------------
 """
import os
import sys

import Cython.Compiler.Options
import numpy as np
from Cython.Distutils import build_ext

try:
    from setuptools import setup
    from setuptools import Extension
except ImportError:
    from distutils.core import setup
    from distutils.extension import Extension

sys.dont_write_bytecode = True

Cython.Compiler.Options.annotate = True

here = os.path.dirname(os.path.realpath(__file__))
whole_path = os.path.join(here, "AoN.pyx")

ext_module = Extension(
    "AoN",
    [whole_path],
    # extra_compile_args=['/fopenmp'],
    # extra_link_args=['/fopenmp'],
    include_dirs=[np.get_include()],
)

setup(cmdclass={"build_ext": build_ext}, ext_modules=[ext_module])



from unittest import TestCase
from aequilibrae.matrix import AequilibraEData, AequilibraeMatrix
from aequilibrae.distribution import SyntheticGravityModel, GravityApplication
import numpy as np
import tempfile
import os

zones = 10

# row vector
args = {
    "entries": zones,
    "field_names": [u"rows"],
    "data_types": [np.float64],
    "memory_mode": True,
}

row_vector = AequilibraEData()
row_vector.create_empty(**args)
row_vector.index[:] = np.arange(row_vector.entries) + 100
row_vector.rows[:] = row_vector.index[:] + np.random.rand(zones)[:]

# column vector
args["field_names"] = ["columns"]
column_vector = AequilibraEData()
column_vector.create_empty(**args)
column_vector.index[:] = np.arange(column_vector.entries) + 100
column_vector.columns[:] = column_vector.index[:] + np.random.rand(zones)[:]

# balance vectors
column_vector.columns[:] = column_vector.columns[:] * (
    row_vector.rows.sum() / column_vector.columns.sum()
)

# Impedance matrix_procedures
name_test = os.path.join(tempfile.gettempdir(), "aequilibrae_matrix_test.aem")

args = {"file_name": name_test, "zones": zones, "matrix_names": ["impedance"]}

matrix = AequilibraeMatrix()
matrix.create_empty(**args)

# randoms = np.random.randint(5, size=(2, 4))
matrix.impedance[:, :] = np.random.rand(zones, zones)[:, :]
matrix.index[:] = np.arange(matrix.zones) + 100
matrix.computational_view(["impedance"])

model_expo = SyntheticGravityModel()
model_expo.function = "EXPO"
model_expo.beta = 0.1

model_gamma = SyntheticGravityModel()
model_gamma.function = "GAMMA"
model_gamma.beta = 0.1
model_gamma.alpha = -0.2

model_power = SyntheticGravityModel()
model_power.function = "POWER"
model_power.alpha = -0.2


class TestGravityApplication(TestCase):
    def test_apply(self):
        args = {
            "impedance": matrix,
            "rows": row_vector,
            "row_field": "rows",
            "columns": column_vector,
            "column_field": "columns",
        }

        models = [("EXPO", model_expo), ("POWER", model_power), ("GAMMA", model_gamma)]

        for model_name, model_obj in models:
            args["model"] = model_obj
            distributed_matrix = GravityApplication(**args)
            distributed_matrix.apply()

            if (
                distributed_matrix.gap
                > distributed_matrix.parameters["convergence level"]
            ):
                self.fail(
                    "Gravity application did not converge for model " + model_name
                )



import tempfile
from os.path import dirname, abspath, join

# For the graph tests
test_network = join(dirname(dirname(abspath(__file__))), "data", "Final_Network.shp")
test_graph = join(dirname(dirname(abspath(__file__))), "data", "test_graph.aeg")
path_test = tempfile.gettempdir()

gtfs_folder = join(dirname(dirname(abspath(__file__))), "data/gtfs")
gtfs_zip = join(dirname(dirname(abspath(__file__))), "data/gtfs.zip")
gtfs_db_output = join(path_test, "test.db")

# For the skimming test


# For the matrix test
omx_example = join(dirname(dirname(abspath(__file__))), "data/test_omx.omx")



import sys
import os
import numpy as np
from setuptools import setup, find_packages
from setuptools import Extension
from Cython.Distutils import build_ext
from aequilibrae.paths.__version__ import release_version

sys.dont_write_bytecode = True

here = os.path.dirname(os.path.realpath(__file__))
whole_path = os.path.join(here, "aequilibrae/paths", "AoN.pyx")
ext_module = Extension("aequilibrae.paths.AoN", [whole_path], include_dirs=[np.get_include()])

pkgs = [pkg for pkg in find_packages()]

pkg_data = {
    "aequilibrae.reference_files": ["spatialite.sqlite"],
    "aequilibrae.paths": ["parameters.pxi"],
    "aequilibrae": ["parameter_default.yml", "parameters.yml"],
}
loose_modules = ["__version__", "parameters", "reserved_fields"]


if __name__ == "__main__":
    setup(
        name="aequilibrae",
        version=release_version,
        install_requires=["numpy", "PyQt5", "pyaml"],
        packages=pkgs,
        package_dir={"": "."},
        py_modules=loose_modules,
        package_data=pkg_data,
        zip_safe=False,
        description="A package for transportation modeling",
        author="Pedro Camargo",
        author_email="pedro@xl-optim.com",
        url="https://github.com/AequilibraE/aequilibrae",
        license="See license.txt",
        classifiers=[
            "Programming Language :: Python",
            "Programming Language :: Python :: 3.5",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
        ],
        cmdclass={"build_ext": build_ext},
        ext_modules=[ext_module],
    )



import numpy as np


class MultiThreadedNetworkSkimming:
    def __init__(self):
        # The predecessors for each node in the graph
        self.predecessors = None
        # holds the skims for all nodes in the network (during path finding)
        self.temporary_skims = None
        # Keeps the order in which the nodes were reached for the cascading network loading
        self.reached_first = None
        # The previous link for each node in the tree
        self.connectors = None
        #  holds the b_nodes in case of flows through centroid connectors are blocked
        self.temp_b_nodes = None

    # In case we want to do by hand, we can prepare each method individually
    def prepare(self, graph, results):
        itype = graph.default_types("int")
        ftype = graph.default_types("float")
        self.predecessors = np.zeros((results.nodes, results.cores), dtype=itype)
        self.temporary_skims = np.zeros((results.nodes, results.num_skims, results.cores), dtype=ftype)
        self.reached_first = np.zeros((results.nodes, results.cores), dtype=itype)
        self.connectors = np.zeros((results.nodes, results.cores), dtype=itype)
        self.temp_b_nodes = np.zeros((graph.b_node.shape[0], results.cores), dtype=itype)

        for i in range(results.cores):
            self.temp_b_nodes[:, i] = graph.b_node[:]



import datetime
import os
import random
import openmatrix as omx
from unittest import TestCase

import numpy as np
from tables.exceptions import NoSuchNodeError

from aequilibrae.matrix import AequilibraeMatrix
from ...data import omx_example

zones = 50
name_test = AequilibraeMatrix().random_name()
copy_matrix_name = AequilibraeMatrix().random_name()
csv_export_name = copy_matrix_name + ".csv"
omx_export_name = copy_matrix_name + ".omx"


class TestAequilibraeMatrix(TestCase):
    def test___init__(self):
        os.remove(name_test) if os.path.exists(name_test) else None
        args = {
            "file_name": name_test,
            "zones": zones,
            "matrix_names": ["mat", "seed", "dist"],
            "index_names": ["my indices"],
        }

        matrix = AequilibraeMatrix()
        matrix.create_empty(**args)

        matrix.index[:] = np.arange(matrix.zones) + 100
        matrix.mat[:, :] = np.random.rand(matrix.zones, matrix.zones)[:, :]
        matrix.mat[:, :] = matrix.mat[:, :] * (1000 / np.sum(matrix.mat[:, :]))
        matrix.setName("Test matrix - " + str(random.randint(1, 10)))
        matrix.setDescription("Generated at " + datetime.datetime.now().strftime("%I:%M%p on %B %d, %Y"))
        matrix.close()
        del matrix

    def test_load(self):
        # self.test___init__()
        self.new_matrix = AequilibraeMatrix()
        self.new_matrix.load(name_test)

    def test_computational_view(self):
        self.test_load()
        self.new_matrix.computational_view(["mat", "seed"])
        self.new_matrix.mat.fill(0)
        self.new_matrix.seed.fill(0)
        if self.new_matrix.matrix_view.shape[2] != 2:
            self.fail("Computational view returns the wrong number of matrices")

        self.new_matrix.computational_view(["mat"])
        self.new_matrix.matrix_view[:, :] = np.arange(zones ** 2).reshape(zones, zones)
        if np.sum(self.new_matrix.mat) != np.sum(self.new_matrix.matrix_view):
            self.fail("Assigning to matrix view did not work")
        self.new_matrix.setName("Test matrix - " + str(random.randint(1, 10)))
        self.new_matrix.setDescription("Generated at " + datetime.datetime.now().strftime("%I:%M%p on %B %d, %Y"))
        self.new_matrix.close()

    def test_copy(self):
        self.test_load()

        # test in-memory matrix_procedures copy

        matrix_copy = self.new_matrix.copy(copy_matrix_name, cores=["mat"])

        if not np.array_equal(matrix_copy.mat, self.new_matrix.mat):
            self.fail("Matrix copy was not perfect")
        matrix_copy.close()
        self.new_matrix.close()

    def test_export_to_csv(self):
        self.test_load()
        self.new_matrix.export(csv_export_name)
        self.new_matrix.close()

    def test_export_to_omx(self):
        self.test_load()
        self.new_matrix.export(omx_export_name)

        omxfile = omx.open_file(omx_export_name, "r")

        # Check if matrices values are compatible
        for m in self.new_matrix.names:
            sm = np.nansum(self.new_matrix.matrix[m])
            sm2 = np.nansum(np.array(omxfile[m]))
            if sm != sm2:
                self.fail("Matrix {} was exported with the wrong value".format(m))
        self.new_matrix.close()

    def test_nan_to_num(self):
        self.test_load()
        s = self.new_matrix.seed.sum() - self.new_matrix.seed[1, 1]
        m = self.new_matrix.mat.sum() - self.new_matrix.mat[1, 1]
        self.new_matrix.seed[1, 1] = np.nan
        self.new_matrix.computational_view(["mat", "seed"])
        self.new_matrix.nan_to_num()
        self.new_matrix.mat[1, 1] = np.nan
        self.new_matrix.computational_view(["mat"])
        self.new_matrix.nan_to_num()

        if s != self.new_matrix.seed.sum():
            self.fail("Total for seed matrix not maintained")

        if m != self.new_matrix.mat.sum():
            self.fail("Total for mat matrix not maintained")

    def test_copy_from_omx(self):
        temp_file = AequilibraeMatrix().random_name()
        a = AequilibraeMatrix()
        a.create_from_omx(temp_file, omx_example)

        omxfile = omx.open_file(omx_example, "r")

        # Check if matrices values are compatible
        for m in ["m1", "m2", "m3"]:
            sm = a.matrix[m].sum()
            sm2 = np.array(omxfile[m]).sum()
            if sm != sm2:
                self.fail("Matrix {} was copied with the wrong value".format(m))

        if np.any(a.index[:] != np.arange(a.zones)):
            self.fail("Index was not created properly")
        a.close()

    def test_copy_from_omx_long_name(self):

        temp_file = AequilibraeMatrix().random_name()
        a = AequilibraeMatrix()

        with self.assertRaises(ValueError):
            a.create_from_omx(temp_file, omx_example, robust=False)

    def test_copy_omx_wrong_content(self):
        # Check if we get a result if we try to copy non-existing cores
        temp_file = AequilibraeMatrix().random_name()
        a = AequilibraeMatrix()

        with self.assertRaises(ValueError):
            a.create_from_omx(temp_file, omx_example, cores=["m1", "m2", "m3", "m4"])

        with self.assertRaises(ValueError):
            a.create_from_omx(temp_file, omx_example, mappings=["wrong index"])



from unittest import TestCase

from aequilibrae.transit.gtfs import GTFS

# Adds the folder with the data to the path and collects the paths to the files
from ...data import gtfs_folder


class TestGTFS(TestCase):
    def setUp(self):
        self.gtfs = GTFS()
        self.gtfs.source_folder = gtfs_folder

    def test_load_calendar_dates(self):
        self.gtfs.load_calendar_dates()
        if self.gtfs.schedule_exceptions != set(["FULLW"]):
            self.fail("calendar_dates.txt was read wrong")

    def test_load_agency(self):
        try:
            self.gtfs.load_agency()
        except Exception as err:
            self.fail("Agency loader returned an error - {}".format(err.__str__()))
        if self.gtfs.agency.name != "Public Transport":
            self.fail("Agency name was read wrong")

    def test_load_stops(self):
        try:
            self.gtfs.load_stops()
        except Exception as err:
            self.fail("stops loader returned an error - {}".format(err.__str__()))
        self.assertEqual(len(self.gtfs.stops), 88, "Not all values read")
        if self.gtfs.stops["88"].name != "Post Office":
            self.fail("GTFS stops not read properly")

    def test_load_routes(self):
        self.gtfs.load_routes()
        if self.gtfs.routes["1415"].long_name != "Braitling and Ciccone":
            self.fail("Route long name not read properly")

        if self.gtfs.routes["1825"].short_name != "500":
            self.fail("Route long name not read properly")

    def test_load_trips(self):
        self.gtfs.load_trips()
        # self.fail()

    def test_load_shapes(self):
        self.gtfs.load_shapes()

    def test_get_routes_shapes(self):
        self.gtfs.load_trips()
        self.gtfs.load_routes()
        self.gtfs.load_shapes()

    def get_data(self):
        self.gtfs = GTFS()
        self.gtfs.source_folder = gtfs_folder



import os
import sys
from unittest import TestCase

from aequilibrae.paths import path_computation, Graph
from aequilibrae.paths.results import PathResults
from ...data import test_graph

# Adds the folder with the data to the path and collects the paths to the files
lib_path = os.path.abspath(os.path.join("..", "../tests"))
sys.path.append(lib_path)

origin = 5
dest = 27


class TestPathResults(TestCase):
    def test_prepare(self):
        # graph
        self.g = Graph()
        self.g.load_from_disk(test_graph)
        self.g.set_graph(cost_field="distance", skim_fields=None)

        self.r = PathResults()
        try:
            self.r.prepare(self.g)
        except Exception as err:
            self.fail("Path result preparation failed - {}".format(err.__str__()))

    def test_reset(self):
        self.test_prepare()
        try:
            self.r.reset()
        except Exception as err:
            self.fail("Path result resetting failed - {}".format(err.__str__()))

    def test_update_trace(self):
        self.test_prepare()
        try:
            self.r.reset()
        except Exception as err:
            self.fail("Path result resetting failed - {}".format(err.__str__()))

        path_computation(origin, dest, self.g, self.r)

        if list(self.r.path) != [53, 52, 13]:
            self.fail("Path computation failed. Wrong sequence of links")

        if list(self.r.path_nodes) != [5, 168, 166, 27]:
            self.fail("Path computation failed. Wrong sequence of path nodes")

        if list(self.r.milepost) != [0, 341, 1398, 2162]:
            self.fail("Path computation failed. Wrong milepost results")
